header "pager" /{
/*
 * Page directory structure for fast message identification
 * ACKNOWLEDGEMENTS to roland urban roland.urban@ipk.fhg.de who provided 
 * me with this algorithm
 *
 */

typedef unsigned int   UINT;
typedef unsigned short USHORT;
typedef unsigned long  ULONG;
typedef unsigned char  BYTE;


typedef struct {
	ULONG  structsize;       /* size of the paging memory */
	USHORT granularity;      /* number of entrys in a page directory */
        USHORT msgIDlength;      /* number of bits in a Message ID */
        ULONG  mask;             /* higher sign. bit mask */
        USHORT nPages;           /* number of reserved pages */
        USHORT nUsed;            /* number of used pages */
} pagerConfig;

typedef struct {
  pagerConfig      *pConfig;
  USHORT *pPages;
  canmsg_t       *pEntry;
} msgPager;


extern pagerConfig *Global_pConfig[];
extern msgPager     Global_msgPager[];

}/

file "pager" /{
/*
 * initialize Pager structure:
 *
 */

pagerConfig *Global_pConfig[4] = { NULL, NULL, NULL, NULL };
msgPager     Global_msgPager[4];

}/


/@
   Initializes a paged memory area of nPages;
       the granularity is the number of bits in the message ID that should be paged
       for example a granularity of 4 provides 16 page directory entrys. the binary 
       search so reduces to 17 steps.
       msgIDlength is the number of bits in the message ID.

       The number of Pages must be choosen to a sufficient amount to keep all messages
       in a block of 16 msg IDs.

@/
func returns pagerConfig* initPager ( USHORT granul, USHORT msgIDlength, USHORT nPages ) /{
        ULONG      required;
	pagerConfig * pConfig;


	required =  sizeof(pagerConfig);
	required += sizeof(USHORT) * (1<<(msgIDlength-granul));
	required += sizeof(canmsg_t) * (1<<granul)*nPages;
	
        if((pConfig = (pagerConfig *) kmalloc(required,GFP_KERNEL))==NULL){
		DBGprint(DBG_DATA,("memory problem"));
		DBGout();return NULL;
	}
	memset((pagerConfig *) pConfig, 0,required);	

	pConfig->structsize  = required;
	pConfig->granularity = granul;
        pConfig->msgIDlength = msgIDlength;
        pConfig->mask        = (1<<granul)-1;
	pConfig->nUsed	     = 0;
	

	DBGout();return pConfig;
}/


/@
   Configures the pager structs
 @/
func confPager ( pagerConfig* pConfig, msgPager* pPager ) /{
        void* p = pPager->pConfig = pConfig;

        pPager->pPages  = (USHORT*)     p += sizeof(pagerConfig);
        pPager->pEntry  = (canmsg_t*)   p += pConfig->nPages * sizeof(USHORT);
}/


/@
  Frees the page structures
 @/
func freePager ( pagerConfig *pConfig ) /{
	kfree((void *) pConfig);
}/

/@
  Binds a Message ID to a page, prepares and allocates a page for a given ID
@/

func returns ULONG bindRTRMsg( msgPager* pPager, ULONG msgID ) /{

    ULONG       page;
    ULONG       index;
        
        page = msgID >> pPager->pConfig->granularity;

        /*      If page is not availiable get new one from Pool
                or return na error
        */
        if( !pPager->pPages[page] ) {
                if( pPager->pConfig->nUsed == pPager->pConfig->nPages ){
			DBGprint(DBG_DATA,("failed"));
                        DBGout();return -1;
		}
                pPager->pPages[page] = ++(pPager->pConfig->nUsed);
        }

        /*      Calculate message index:
                - Page-Index * Granularity (=Basis)
                - Rest-Index from msgID (+Offset)
        */
        index  = (pPager->pPages[page]-1) << pPager->pConfig->granularity;
        index += msgID & pPager->pConfig->mask;

        DBGout();return index;
}/


/@
	Get a message Index for a given ID
@/

func returns ULONG findRTRMsg( msgPager* pPager, ULONG msgID ) /{

    ULONG       page;
    ULONG       index;
        
        /*     Calculate Page index
        */
        page = msgID >> pPager->pConfig->granularity;

        /*     No page availiable ?
        */
	DBGprint(DBG_DATA,("id=%d",msgID));
        if( !pPager->pPages[page] ){
		DBGprint(DBG_DATA,("failed"));
                DBGout();return -1;
	}
        
        /*      Calculate Page index:
                - Page-Index * Granularity (=Basis)
                - Rest-Index from msgID (+Offset)
        */
        index  = (pPager->pPages[page]-1)<<pPager->pConfig->granularity;
        index += msgID & pPager->pConfig->mask;

        DBGout();return index;
}/

/@
    Assigns a message to the given ID
@/

func returns USHORT setRTRMsg( msgPager* pPager, canmsg_t *msg ) /{

        USHORT          i;
        ULONG           index;
        canmsg_t*       pEntry;

        /*      
                Get Message for the given ID
        */
        if((index=Can_findRTRMsg(pPager, msg->id ) < 0) ){
                DBGout();return -1;
        }
        pEntry = &(pPager->pEntry[index]);
	DBGprint(DBG_DATA,("id=%d",msg->id));


	memcpy(pEntry,msg,sizeof(canmsg_t));
        
        DBGout();return (pEntry->length=i);
}/


/@
	get a message from pager memory
@/
func returns USHORT getRTRMsg( msgPager* pPager, canmsg_t *msg ) /{

        USHORT          i;
        ULONG           index;
        canmsg_t*       pEntry;

        if( (index=Can_findRTRMsg(pPager,msg->id)) <0 ){
                DBGout();return -1;
	}

        pEntry = &(pPager->pEntry[index]);
	DBGprint(DBG_DATA,("id=%d",msg->id));


	memcpy(msg,pEntry,sizeof(canmsg_t));


        DBGout();return msg->length;
}/


